# -*- coding: utf-8 -*-
# OMG Music Manager  -  http://omg.mathematik.uni-kl.de
# Copyright (C) 2009-2013 Martin Altmayer, Michael Helmling
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from PyQt4 import QtGui

from .. import application, constants, database as db, logging, utils
from ..constants import ADDED, DELETED, CHANGED, ADD, DELETE, CHANGE
from ..application import ChangeEvent


logger = logging.getLogger(__name__)
translate = QtGui.QApplication.translate

_flagsById = None
_flagsByName = None


def init():
    """Initialize the flag module loading flags from the database. You must call this method before methods
    like get can be used."""
    global _flagsById, _flagsByName, flagList
    _flagsById = {}
    _flagsByName = {}
    for row in db.query("SELECT id,name,icon FROM {}flag_names".format(db.prefix)):
        id,name,iconPath = row
        if db.isNull(iconPath):
            iconPath = None
        flagType = Flag(id,name,iconPath)
        _flagsById[flagType.id] = flagType
        _flagsByName[flagType.name] = flagType


class Flag:
    """A flagtype with an id, a name and optionally an icon. At first glance flags are like tags, but in fact
    they are much easier, because they have no values, valuetypes, translations and because they are not
    stored in files.
    
    Attributes of Flag: id, name, iconPath
    Read-only attribute: icon
    
    Usually you shold get flag instances via flags.get. The exception is for  flags that are not (yet) in
    the database (use :func:`exists` to check this). For these flags get will fail and you have to create
    your own instances. If you use the common instance, it will get automatically updated on
    FlagTypeChangeEvents.
    """
    def __init__(self,id=None,name=None,iconPath=None):
        self.id = id
        self.name = name
        self.iconPath = iconPath
        
    @property
    def iconPath(self):
        return self._iconPath

    @iconPath.setter
    def iconPath(self,iconPath):
        """Set the flag's iconPath and load the icon."""
        self._iconPath = iconPath
        if iconPath is not None:
            self.icon = QtGui.QIcon(iconPath)
        else: self.icon = None
        
    def __str__(self):
        return self.name
    
    def __eq__(self,other):
        return isinstance(other,Flag) and self.id == other.id
    
    def __ne__(self,other):
        return not isinstance(other,Flag) or self.id != other.id
    
    def __hash__(self):
        return self.id
        

def get(identifier):
    """Return a flagtype. *identifier* may be an int (the flag's id), a string (its name) or a flagtype (in
    this case it is simply returned)."""
    if isinstance(identifier,int):
        return _flagsById[identifier]
    elif isinstance(identifier,str):
        return _flagsByName[identifier]
    elif isinstance(identifier,Flag):
        return identifier
    else: raise ValueError("identifier must be either int or string or FlagType.")


def exists(name):
    """Return whether a flagtype with the given name exists."""
    return name in _flagsByName


def isValidFlagname(name):
    """Return whether *name* is a valid name for a flagtype."""
    return 0 < len(name.encode()) <= constants.FLAG_VARCHAR_LENGTH and not name.isspace()


def allFlags():
    """Return a list containing all flags in the database."""
    return _flagsById.values()


def addFlagType(name,**data):
    """Add a new flagtype with the given name to the database. *data* may be used to set attributes. 
    Currently only 'iconPath' is supported."""
    if exists(name):
        raise ValueError("There is already a flag with name '{}'.".format(name))
    if not isValidFlagname(name):
        raise ValueError("'{}' is not a valid flagname.".format(name))
    command = FlagTypeUndoCommand(ADD,name=name,**data)
    application.stack.push(command)
    return command.flagType
    
    
def _addFlagType(**data):
    """Like addFlagType, but not undoable. The keyword arguments may contain either
    
        - a single argument 'flagType': In this case the given flagType is inserted into the database and
          some internal lists. Use this to undo a flagType's deletion.
        - a 'name' and optionally an 'iconPath'
          
    After creation the a FlagTypeChangedEvent is emitted.
    """
    if 'flagType' in data:
        flagType = data['flagType']
        data = (flagType.id,flagType.name,flagType.iconPath)
        db.query(
            "INSERT INTO {}flag_names (id,name,icon) VALUES (?,?,?)"
              .format(db.prefix),*data)
    else:
        # The difference to the if-part is that we have to get the id from the database
        flagType = Flag(**data)
        data = (flagType.name,flagType.iconPath)
        flagType.id = db.query(
            "INSERT INTO {}flag_names (name,icon) VALUES (?,?)"
              .format(db.prefix),*data).insertId()
    logger.info("Added new flag '{}'".format(flagType.name))
    
    _flagsById[flagType.id] = flagType
    _flagsByName[flagType.name] = flagType
    application.dispatcher.emit(FlagTypeChangedEvent(ADDED,flagType))
    return flagType


def deleteFlagType(flagType):
    """Delete a flagtype from the database."""
    application.stack.push(FlagTypeUndoCommand(DELETE,flagType))
    
    
def _deleteFlagType(flagType):
    """Like deleteFlagType but not undoable."""
    if not exists(flagType.name):
        raise ValueError("Cannot remove flagtype '{}' because it does not exist.".format(flagType))
    
    logger.info("Removing flag '{}'.".format(flagType))
    db.query("DELETE FROM {}flag_names WHERE id = ?".format(db.prefix),flagType.id)
    del _flagsById[flagType.id]
    del _flagsByName[flagType.name]
    application.dispatcher.emit(FlagTypeChangedEvent(DELETED,flagType))


def changeFlagType(flagType,**data):
    """Change a flagtype. The attributes that should be changed must be specified by keyword arguments.
    Supported are 'name' and 'iconPath':

        changeFlagType(flagType,name='Great',iconPath=None)
    
    """
    application.stack.push(FlagTypeUndoCommand(CHANGE,flagType,**data))
    
    
def _changeFlagType(flagType,**data):
    """Like changeFlagType but not undoable."""
    # Below we will build a query like UPDATE flag_names SET ... using the list of assignments (e.g. (name=?).
    # The parameters will be sent with the query to replace the questionmarks.
    assignments = []
    params = []
    
    if 'name' in data:
        name = data['name']
        if name != flagType.name:
            if exists(name):
                raise ValueError("There is already a flag named '{}'.".format(name))
            logger.info("Changing flag name '{}' to '{}'.".format(flagType.name,name))
            assignments.append('name = ?')
            params.append(name)
            del _flagsByName[flagType.name]
            _flagsByName[name] = flagType
            flagType.name = name
        
    if 'iconPath' in data and data['iconPath'] != flagType.iconPath:
        assignments.append('icon = ?')
        params.append(data['iconPath'])
        flagType.iconPath = data['iconPath']
    
    if len(assignments) > 0:
        params.append(flagType.id) # for the where clause
        db.query("UPDATE {}flag_names SET {} WHERE id = ?".format(db.prefix,','.join(assignments)),*params)
        application.dispatcher.emit(FlagTypeChangedEvent(CHANGED,flagType))


class FlagTypeUndoCommand:
    """This command adds, changes or deletes a flagtype. Which keyword arguments are necessary depends on the
    first argument *action*. Use the methods addFlagType, deleteFlagType and changeFlagType instead of
    creating a command directly.
    """
    def __init__(self,action,flagType=None,**data):
        self.text = {ADD:   translate("FlagTypeUndoCommand","Add flagType"),
                     DELETE: translate("FlagTypeUndoCommand","Delete flagType"),
                     CHANGE: translate("FlagTypeUndoCommand","Change flagType")
                    }[action]
        self.action = action
        if self.action == ADD:
            self.addData = data
            self.flagType = None
        elif self.action == DELETE:
            self.flagType = flagType
        else:
            self.flagType = flagType
            self.oldData = {'name': flagType.name,'iconPath': flagType.iconPath}
            self.newData = data
        
    def redo(self):
        if self.action == ADD:
            if self.flagType is None: # This is the first time this command is redone
                self.flagType = _addFlagType(**self.addData)
                del self.addData
            else:
                # On subsequent redos ensure that the same object is recreated,
                # because it might be used in many elements within the undohistory.
                _addFlagType(flagType=self.flagType)
        elif self.action == DELETE:
            _deleteFlagType(self.flagType)
        else: _changeFlagType(self.flagType,**self.newData)

    def undo(self):
        if self.action == ADD:
            _deleteFlagType(self.flagType)
        elif self.action == DELETE:
            # Ensure that the same object is recreated, because it might be used in many elements
            # within the undohistory.
            _addFlagType(flagType=self.flagType)
        else: _changeFlagType(self.flagType,**self.oldData)


class FlagTypeChangedEvent(ChangeEvent):
    """FlagTypeChangedEvent are used when a flagtype is added, changed or deleted."""
    def __init__(self,action,flagType):
        assert action in constants.CHANGE_TYPES
        self.action = action
        self.flagType = flagType


class FlagDifference:
    """Stores changes to flag lists and provides methods to apply them. When this difference is applied to
    an element the flags in *removals* will be removed from the elements flags and the flags in *additions*
    will be added.
    """
    def __init__(self, additions=None, removals=None):
        self._additions = additions
        self._removals = removals
        
    def apply(self, element):
        """Change the flags of *element* according to this difference object."""
        if self._removals is not None:
            for flag in self._removals:
                element.flags.remove(flag)
        if self._additions is not None:
            element.flags.extend(self._additions)

    def revert(self, element):
        """Undo the changes of this difference object to the flags of *element*."""
        if self._additions is not None:
            for flag in self._additions:
                element.flags.remove(flag)
        if self._removals is not None:
            element.flags.extend(self._removals)
            
    def getAdditions(self):
        """Return the list of flags which are added by this difference."""
        return self._additions if self._additions is not None else []
        
    def getRemovals(self):
        """Return the list of flags which are removed by this difference."""
        return self._removals if self._removals is not None else []
        
    def inverse(self):
        """Return the inverse difference."""
        return utils.InverseDifference(self)
            

class FlagListDifference(FlagDifference):
    """Subclass of FlagDifference that simply takes two lists of flags (old and new) and figures out
    additions/removals by itself."""
    def __init__(self,oldFlags,newFlags):
        self.oldFlags = oldFlags
        self.newFlags = newFlags
                
    def apply(self, element):
        element.flags = self.newFlags[:]
        
    def revert(self, element):
        element.flags = self.oldFlags[:]
        
    def getAdditions(self):
        return [flag for flag in self.newFlags if not flag in self.oldFlags]
    
    def getRemovals(self):
        return [flag for flag in self.oldFlags if not flag in self.newFlags]
        